/***************************************************************
* ccode:   Chimera and Cross-Over Detection and Evaluation 
* Reads sequence/s in fasta format from a single file. The query sequence
* must be the first sequence followed by the most similar ones. 
*
* (C) Copyright. All rights reserved. Juan M. Gonzalez. 2004.
* Permission to use, reproduce, and distribute this program
* or its parts must be granted by the author.
*
* Date: February 14th, 2004. Last modified: May 13th, 2004
*
* Author: Juan M. Gonzalez
*
* Address: 
*          IRNAS-CSIC
*          P.O.Box 1052,
*	   41080 - Sevilla, Spain
* e-mail:  jmgrau@irnase.csic.es
*
* Build: gcc -o ccode025 ccode025.c -lm 
*
*****************************************************************/
#include <stdio.h>
#include <stdlib.h>     /* standard ANSI only */
#include <string.h>
#include <math.h>

/*---------------------------------------------------*/

#define MAX_LINE 501	/* maximum length of an input line of data */
#define MAX_BASES 4000	/* maximum number of bases per sequence */
#define MAX_SEQS 30	/* maximum number of sequences to be processed */
#define MAX_WORDS 400	/* maximum number of fragments or words to divide a sequence */
#define MIN_DIFF 0.5	/* minimum difference (%) accepted between two reference sequences */
#define MAX_DIFF 20.0	/* maximum difference (%) accepted between two reference sequences */
#define E 0.001		/* minimum error, just to avoid divisions by zero */

/*---------------------------------------------------*/

main(int argc, char *argv[])
{

        /* variables */
        char *program_name;     /* name of the program collected from
                                   the command line (for errors?) */
        char file_in[25];       /* name of the file with the query and related sequence/s */
 	char file_out[25];      /* file name to output the results */

        FILE *data_file_in;     /* file to read the query and related sequences */
    	FILE *data_file_out;    /* file to write the data to */

        char line[MAX_LINE];    /* array to read each input line */ 
	char line_command[MAX_LINE]; /* holds command line */
	char *data_ptr;	        /* character returned by fgets. NULL= end of file */ 
	char seq[MAX_SEQS][MAX_BASES]; /* helds each sequence. Query sequence is the first */
	int bases[MAX_SEQS];	/* counts of the number of bases in each sequence */
	float percent_word;	/* percentage to calculate word size as a funtion of sequence length */
	int word;		/* word size */
	int filter;		/* chooses a filter to apply to the comparisons */
	int sequences = -1;	/* number of sequences read from the input file */
	int no_cr;		/* print copyright information */
	int i;			/* counter for number of sequences */
	int n;			/* counter for number of words */
	int c;			/* counter bases per full sequence */
	int k;			/* counter No. bases in a sequence */
	int w;			/* counter No. bases per word */
	int s, j;		/* counters reading and comparing sequences */
	char name_seq[MAX_SEQS][MAX_LINE]; /* name of sequences. Query sequence will be the first */
        int diffs;              /* number of differences between fragments of sequences */
	float x[MAX_SEQS][MAX_WORDS];	/* differences per word length for the query sequence */
	float z[MAX_SEQS-1][MAX_SEQS-1][MAX_WORDS];/* diffs per word only for related sequences */
	int bx[MAX_SEQS][MAX_WORDS];	/* holds the number of bases in each word */
        int bz[MAX_SEQS-1][MAX_SEQS-1][MAX_WORDS];  /* holds the number of bases in each word */
	int nx[MAX_SEQS];	/* number of words per sequence */
	float avg_x[MAX_WORDS]; /* average for differences per word */
	float sd_x[MAX_WORDS];  /* sd for differences per word */
	float var_x[MAX_WORDS]; /* variance of query with respect to other sequences */
	float avg_z[MAX_WORDS]; /* averages for differences per word for related sequences only */
	float sd_z[MAX_WORDS];  /* sd for differences per word for related sequences only */
	float var_z[MAX_WORDS]; /* variance among reference sequences */
	float sum[MAX_WORDS];	/* cummulative differences per word for current sequence */
	float sq_sum[MAX_WORDS]; /* sum of squares for sum */
	int distinct_no_bases;	/* indicates if  the number of bases is the same in every sequence */
	float ds[MAX_WORDS];	/* distance score: query distances divided by references dist., X/Z */
	float ds1[MAX_WORDS];   /* distance score + sd_z */
	float ds2[MAX_WORDS];	/* distance score - sd_z */
	float fs[MAX_WORDS];    /* F-Snedecor, test for differences of variances */
	int chimera[MAX_WORDS]; /* suggests chimera based on the confidence limit approach */
	int chimera2[MAX_WORDS]; /* suggests chimera based on the T-student test approach */
	int avg_nx;		/* average number of words per sequence */
	int ref_combs;		/* number of combinations between reference sequences */
	float t;		/* value of t-student */
	int df;			/* degrees of freedom for T-student test */
	float denom_ts;		/* denominator, without sqrt(), for ts calculations */
	float ts[MAX_WORDS];	/* value of ts for t-student test */
	float anova[MAX_WORDS]; /* value of anova test of significance */
	float sum_x[MAX_WORDS]; /* holds the value of sums from query vs. reference comparisons */
	float sum_z[MAX_WORDS]; /* holds the value of sums from reference comparisons */
	float sq_sum_x[MAX_WORDS];  /* holds sum squares for query-references */
	float sq_sum_z[MAX_WORDS];  /* holds sum squares for reference comparisons */
	float value1[MAX_WORDS]; /* intermediate value for computing anova */
        float value2[MAX_WORDS]; /* intermediate value for computing anova */
	float value3[MAX_WORDS]; /* intermediate value for computing anova */
	float value4[MAX_WORDS]; /* intermediate value for computing anova */
	float value5[MAX_WORDS]; /* intermediate value for computing anova */
	float value6[MAX_WORDS]; /* intermediate value for computing anova */
	float value7[MAX_WORDS]; /* intermediate value for computing anova */
	float value8[MAX_WORDS]; /* intermediate value for computing anova */
	float value9[MAX_WORDS]; /* intermediate value for computing anova */
	float f;		 /* value for Ft (from the F-table */
	int chimera3[MAX_WORDS]; /* suggests a chimera based on the anova approach */
	int distinct;		 /* signals if ambiguous bases are equal (0) or distict (1) */
	float mind;		 /* minimum difference accepted between reference sequences */
	float maxd;		 /* maximum difference accepted between reference sequences */
	int this_one[MAX_SEQS];		 /* sequences to be removed */
	int prediffs[MAX_SEQS][MAX_SEQS];	/* differences among pairs of reference sequences */
	int rem_sim_diff;	/* signals if references sequences should be screened for similarity or
				   differences */
	int num_seqs_left;       /* number of reference sequences left after rem_sim_diff */

                /* declaration of functions */
        void usage(char *program_name);
        void print_cr(void);
	void print_header(FILE *data_file_out_ptr, char *line_command);
	void print_results(FILE *data_file_out_ptr, int word, float avg_x[], float sd_x[], float avg_z[], float sd_z[], float ds[], int chimera[], int chimera2[], int chimera3[], float fs[], float ts[], float anova[], int ref_combs, int sequences, int avg_nx, int bases[], char name_seq[][], int predifs[][]);
	void obtaining_t(float *t_ptr, int df);
	void obtaining_f(float *f_ptr, int v1, int v2);
	void ambiguity(int *different_base_ptr, char base_a, char base_b);
	void remove_extra_gaps(char seq[][], int bases[], int sequences);
 	void sequence_off(char seq[][], int bases[], int *sequences_ptr, int this_one[], char name_seq[][], int prediffs[][]);

/* ---------------------------------------------------------------- */
	/* variables affected by command-line options. Default values */

	no_cr = 0;          /* print copyright information */
	word = -1;  	    /* to check if word length has been given or not */
	percent_word = -1;  /* to check if this data has been input on the command line */
	filter = 0;	    /* No filtering of the sequences */
	mind = MIN_DIFF;    /* accept the default for minimum defference among reference sequences */
	maxd = MAX_DIFF;    /* accept the default for maximum difference among reference sequences */
	rem_sim_diff = 1;   /* eliminate reference sequences too similar or distant */

/* ---------------------------------------------------------------- */

        /* command line arguments stuff */
	(void)strcpy(line_command, " ");
	for (i = 0; i < argc; i++) {
		(void)strcat(line_command, argv[i]);
		(void)strcat(line_command, " ");
	}

        program_name = argv[0];         /* for errors and usage function */

        if (argc == 1) {
                (void)usage(program_name);
                exit(8);
        }
        while ((argc>1) && (argv[1][0] == '-')) {

                switch (argv[1][1]) {

                          /*
                          *  -I<file_name> input file name with sequences 
                          *    argv[1][0] is the dash
                          *    argv[1][1] is the 'I'
                          *    argv[1][2] starts the name
                          */
                        case 'I':
                          (void)strcpy(file_in, &argv[1][2]);
                          break;

                          /*
                          *  -O<file_name> output file name  
                          *    argv[1][0] is the dash
                          *    argv[1][1] is the 'O'
                          *    argv[1][2] starts the name
                          */
                        case 'O':
                          (void)strcpy(file_out, &argv[1][2]);
                          break;

                          /*
                          *  -w word length (integer)  
                          *    argv[1][0] is the dash
                          *    argv[1][1] is the 'w'
                          *    argv[1][2] integer
                          */
                        case 'w':
                          if (word < 0) word = atoi(&argv[1][2]);
                          break;

                          /*
                          *  -p percentage of sequence length to calculate word size  
                          *    argv[1][0] is the dash
                          *    argv[1][1] is the 'p'
                          *    argv[1][2] float number
                          */
                        case 'p':
                          percent_word = atof(&argv[1][2]);
                          break;

                          /*
                          *  -f select filter to use; default = 0, no filter
                          *    argv[1][0] is the dash
                          *    argv[1][1] is the 'f'
                          *    argv[1][2] integer
                          */
                        case 'f':
                          filter = atoi(&argv[1][2]);
                          break;

			  /*
                          *  -M Maximum difference accepted between two reference sequences
                          *    argv[1][0] is the dash
                          *    argv[1][1] is the 'M'
                          *    argv[1][2] float number
                          */
                        case 'M':
                          maxd = atof(&argv[1][2]);
                          break;

		          /*
                          *  -m Minimum difference accepted between two reference sequences
                          *    argv[1][0] is the dash
                          *    argv[1][1] is the 'm'
                          *    argv[1][2] float number
                          */
			case 'm':
			  maxd = atof(&argv[1][2]);
			  break;

                          /*
         	          *  -r select if reference sequences should be screened for high similarity
			  *     or distance
			  *    argv[1][0] is the dash
			  *    argv[1][1] is the 'f'
			  *    argv[1][2] no use of any number 
			  */
	                case 'r':
	                  rem_sim_diff = 0;
	                  break;

                          /*
                          *  -c avoid printing the copyright information
                          *    argv[1][0] is the dash
                          *    argv[1][1] is the 'c'
                          *    argv[1][2] no use of any filename or number
                          */
                        case 'c':
                          no_cr = 1;
                          break;

			  /*
			  *   -h displays help
			  *     argv[1][0] is the dash
			  *     argv[1][1] is the 'h'
			  *     argv[1][2] not used
			  */
			case 'h':
			  (void)usage(program_name);
			  break;

               /* insert other options for the command line here */
 
                        default:
                          (void)fprintf(stderr, "Bad option %s\n", argv[1]);  
                          (void)usage(program_name);
                }  /* end of switch */

                /* move the argument list up one, move the count down one */
                argc--;
                argv++;
        }       /* end of while */

/* ========================================================================== */

	/* print copyright information */
	if (no_cr == 0) {
		(void)print_cr();
	}

        /* initialize the output file */
        data_file_out = fopen(file_out, "w");
	(void)print_header(data_file_out, line_command);

        /* initialize the input files */
	data_file_in = fopen(file_in, "r");
	/* check if the file can be open */
	if (data_file_in == NULL) {
	        (void)fprintf(stderr, "Sequence file for query sequences could not be opened.\n \n");
	        exit(8);
	}

	/* Read query and reference sequences from input file */
	while (1) {

               /* read one line at a time */
               data_ptr = fgets(line, sizeof(line), data_file_in);
               /* check if we have read every line from input */
	       if (data_ptr == NULL) {
	           /* end of file. */
	           /* just do like at the end of every sequence. */
		       break;
	       }
	       else if ((line[0] == '#') || (line[0] == '*')) {
	 	/* This will indicate a comment (#) or end of sequence (in pir format, *)
		 * which are not required. Do nothing */
	       }
	       else if (line[0] == '\n') {
			/* This is an empty line. These appear in the pir format. Do nothing. */
	       }
	       else if (line[0] == '>') {
		       /* This is a header line from the fasta format => Starting a sequence */
                        /* initialize or update counters */
		       sequences = sequences + 1;
		       bases[sequences] = 0;
		       c = 0;

                        /* get sequence name */
		       i = 0;  /* counter for caracters per line. line[0] is the > symbol */
		       while ((line[i] != '\0') && (line[i] != '\n')) {
		             /* look for the end of the first word: space, or period */
		             if ((line[i] == ' ') || (line[i] == '.')) {
		                     i = i - 1;
		                     break;
		             }
		             i = i + 1;
		             line[i-1] = line[i];/* move characters one position lower in the array */
		       }
		       if ((line[i] == '\0') || (line[i] == '\n')) i = i - 1;
		       line[i] = '\0';

		       (void)strcpy(name_seq[sequences], line);

	       }
               else if (line[0] != '>') {
	               s = 0;  /* initialize the counter for bases per line */
	                       /* proccess sequence in current line */
	               while ((line[s] != '\0') && (line[s] != '\n')) {
			       seq[sequences][c] = line[s];
	                       bases[sequences] = bases[sequences] + 1;
			       c = c + 1;
			       s = s + 1;
		       }  /* While */
	       }
	}
	/* we are done reading sequences */
        /* close input file, we are done with it */
        (void)fclose(data_file_in);

	/* Check if every sequence has equal length */
	distinct_no_bases = 0;
	for (i=0; i<sequences; i++) {
		if (bases[i] != bases[i+1]) distinct_no_bases = 1;
	}
	if (distinct_no_bases == 1) {
		(void)fprintf(stdout, "\n \n The input sequences do not all have the same\n");
		(void)fprintf(stdout, "number of bases including the gaps.\n \n \n");
 		exit(8);  /* quit the program */
	}

        /* eliminate initial and final gaps */
        (void)remove_extra_gaps (seq, bases, sequences);

	/* get word size if needed */
	if (word < 0) {
		if (percent_word < 0) {
			(void)fprintf(stdout, "The size or percentage for fragmenting the sequences is required.\n");
			exit(8);
		}
		else {
			word = (percent_word * bases[1])/100;	
		}
	}


        /* Check if reference sequences are within a range of acceptability
	 * according to the differences among them, in order to remove too different
	 * or practically identical sequences */
	if ((rem_sim_diff == 1) && (sequences > 4)) {
	for (i=0; i<sequences+1; i++) {  /* avoid the sequence 0 because it is the query */
	        for (j=1; j<sequences+1; j++) {
	                if (i < j) {
	                        k = 0;
	                        diffs = 0;
				prediffs[i][j] = 0;
				while (k < bases[i]) {
	                                if (seq[i][k] != seq[j][k]) {
	                                        distinct = 1;
	                                        /* filter 0. Default, no filters */
	                                        if (filter == 0) {  /* no filters */
	                                        /* process ambiguous bases */
	                                                (void)ambiguity(&distinct, seq[i][k], seq[j][k]);
                                                        if (distinct == 1)
								diffs = diffs + 1;
                                                }
                                                else if (filter == 1) {
                                                /* filter = 1. Remove all gaps */
                                                     if ((seq[i][k] != '-') && (seq[j][k] != '-')) {
                                                     /* process ambiguous bases */
                                                           (void)ambiguity(&distinct, seq[i][k], seq[j][k]);
				   			   if (distinct == 1)
							          diffs = diffs + 1;
					             }
					        }
					} /* if */
					k = k + 1;
				} /* while */
				prediffs[i][j] = diffs;
			}  /* if */
		}
       	}

	/* Select sequences to be removed and if any, remove them */
	for (j=0; j<sequences+1; j++) this_one[j] = 0; /* initialize variable*/
	for (i=0; i<sequences+1; i++) {  
	        for (j=1; j<sequences+1; j++) {
                        if (i < j) {
				if (prediffs[i][j] < ((mind*1.0*bases[i])/100.0)) {
					this_one[j] = 1;
				}
				if (prediffs[i][j] > ((maxd*1.0*bases[i])/100.0)) {
					this_one[j] = 1;
				}	
			}
		}
	}
	num_seqs_left = 0;  /* count number of sequences left */
	for (j=1; j<sequences+1; j++) {
		if (this_one[j] == 0) num_seqs_left = num_seqs_left + 1;
	}
	if (num_seqs_left >= 3)  /* be sure there are at least three reference sequences left */
		(void)sequence_off(seq, bases, &sequences, this_one, name_seq, prediffs);
	else if (num_seqs_left <= 2) {
		/* Do not remove any sequence */
		(void)fprintf(data_file_out, "Sequence: %s. Problems with reference sequences. Please, take a look.\n ", name_seq[0]);
	}

	/* number of sequences and reference sequences have been updated */

	} /* rem_sim_diff */


	/* Start comparing the reference sequences */
	for (i=1; i<sequences+1; i++) {  /* avoid the sequence 0 because it is the query */
		for (j=1; j<sequences+1; j++) {
			if (i < j) {
				k = 0;
				w = 0;
				n = 0;
				diffs = 0;
				while (k < bases[i]) {
                                        if (seq[i][k] != seq[j][k]) {
						distinct = 1;
                                        	/* filter 0. Default, no filters */
	                                        if (filter == 0) {  /* no filters */
							/* process ambiguous bases */
						        (void)ambiguity(&distinct, seq[i][k], seq[j][k]);
							if (distinct == 1)
	                                                	diffs = diffs + 1; 
						}
						else if (filter == 1) {
					        /* filter = 1. Remove all gaps */
						        if ((seq[i][k] != '-') && (seq[j][k] != '-')) {
							     /* process ambiguous bases */
                                                             (void)ambiguity(&distinct, seq[i][k], seq[j][k]);
                                                             if (distinct == 1)
						                    diffs = diffs + 1;
						        }
						}
		                        }
					k = k + 1;
					w = w + 1;
					if (w >= (word-1)) {
						z[i][j][n] = diffs;
						bz[i][j][n] = w;
						diffs = 0;
						w = 0;
						n = n + 1;
					} /* if word */
				} /* while */
			}  /* if */
		}
	}

	/* Start comparing query sequence with reference sequences */
	for (i=1; i<sequences+1; i++) {
                        k = 0;
                        w = 0;
                        n = 0;
                        diffs = 0;
                        while (k < bases[i]) {
                                if (seq[0][k] != seq[i][k]) {
					distinct = 1;
					/* filter 0. Default, no filters */
					if (filter == 0) {
						/* process ambiguous bases */
                                                (void)ambiguity(&distinct, seq[0][k], seq[i][k]);
                                                if (distinct == 1)
							diffs = diffs + 1;
					}
					else if (filter == 1) {
					/* filter = 1. Remove all gaps */
						if ((seq[0][k] != '-') && (seq[i][k] != '-')) {
							/* process ambiguous bases */
	                                               (void)ambiguity(&distinct, seq[0][k], seq[i][k]);
       		                                       if (distinct == 1)
								diffs = diffs + 1;
						}
					}
				}
                                k = k + 1;
                                w = w + 1;
                                if (w >= (word-1)) {
                                        x[i][n] = diffs;
                                        bx[i][n] = w;
					diffs = 0;
                                        w = 0;
                                        n = n + 1;
                                } /* if word */
                        } /* while */
			nx[i] = n;
        }

	/* Number of words should be the same for every sequence */
	avg_nx = nx[1];

	/* Calculate parameters */
	/* initialize sum and sq_sum */
	for (n=0; n<MAX_WORDS; n++) {
		sum[n] = 0.0;
		sq_sum[n] = 0.0;
	}
	ref_combs = 0;
	/* averages for reference sequences */
	for (i=1; i<sequences+1; i++) {
		for (j=1; j<sequences+1; j++) {
			if (i < j) {
				n = 0;
				ref_combs = ref_combs + 1;
				while (n < avg_nx) {
					sum[n] = sum[n] + z[i][j][n];
					sq_sum[n] = sq_sum[n] + z[i][j][n]*z[i][j][n];
					n = n + 1;
				} 
			}
		}
	}
	for (n=0; n<avg_nx; n++) {
		avg_z[n] = sum[n]/(1.0*ref_combs);  
		var_z[n] = (sq_sum[n] - ((sum[n]*sum[n])/(1.0*ref_combs)))/(1.0*(ref_combs-1));
		sd_z[n] = sqrt(var_z[n]);
		/* hold values of sums for computing anovas */
		sum_z[n] = sum[n];
		sq_sum_z[n] = sq_sum[n];
		/* avoid 0 values */
                if (avg_z[n] < E) avg_z[n] = E;
                if (sd_z[n] < E) sd_z[n] = E;
		if (var_z[n] < E) var_z[n] = E;
		if (sum_z[n] < E) sum_z[n] = E;
		if (sq_sum_z[n] < E) sq_sum_z[n] = E;
	}

        /* initialize sum and sq_sum */
        for (n=0; n<MAX_WORDS; n++) {
                sum[n] = 0.0;
                sq_sum[n] = 0.0;
	}
	/* averages for query sequence */
	for (i=1; i<sequences+1; i++) {
		n = 0;
        	while (n < avg_nx) {
                	sum[n] = sum[n] + x[i][n];
                	sq_sum[n] = sq_sum[n] + x[i][n]*x[i][n];
			n = n + 1;
        	}
	}
        for (n=0; n<avg_nx; n++) {
                avg_x[n] = sum[n]/(1.0*(sequences));    /* assuming No. words is equal for every sequence */
		var_x[n] = (sq_sum[n] - ((sum[n]*sum[n])/(1.0*sequences)))/(1.0*(sequences-1));
		sd_x[n] = sqrt(var_x[n]);
		/* hold values of sums for computing anovas */
		sum_x[n] = sum[n];
		sq_sum_x[n] = sq_sum[n];	
		/* avoid 0 values */
                if (avg_x[n] < E) avg_x[n] = E;
                if (sd_x[n] < E) sd_x[n] = E;
		if (var_x[n] < E) var_x[n] = E;
                if (sum_x[n] < E) sum_x[n] = E;
                if (sq_sum_x[n] < E) sq_sum_x[n] = E;
	}
	/* estimate ratio of variability relative to the variability of each word for that group */
	n = 0;
	while (n < avg_nx) {
		ds[n] = avg_x[n]/(avg_z[n]);

		/* estimating confidence limit for comparisons with query sequence */
		(void)obtaining_t(&t, sequences-1);
		ds1[n] = (avg_x[n] + (t * sd_x[n]))/(avg_z[n]);
                ds2[n] = (avg_x[n] - (t * sd_x[n]))/(avg_z[n]);

		if ((ds1[n] > 1.0) && (ds2[n] > 1.0) && (avg_x[n] > avg_z[n])) {  /* range does not include 1 */
                        chimera[n] = 1;   /* significantly higher at P<0.05 */ 
                }
                else {
                        chimera[n] = 0;
                }

		/* t-student test */
                /* obtaining t-student value and its level of significance */
                df = ref_combs + sequences - 2;
                denom_ts = (((ref_combs - 1)*var_x[n] + (sequences - 1)*var_z[n])/(df*1.0))*((ref_combs + sequences)/(1.0 * ref_combs * sequences));
                ts[n] = fabs((avg_x[n] - avg_z[n])/(sqrt(denom_ts)));
                (void)obtaining_t(&t, df);
		if ((ts[n] >= t) && (avg_x[n] > avg_z[n])) {   
			chimera2[n] = 1;  /* significantly higher at P<0.05 */ 
		}
		else {
			chimera2[n] = 0;
		} 		

		/* F test for differences among variances.
		 * var_x is expected to be higher or similar than var_z */
		fs[n] = var_x[n]/var_z[n];

		/* Anova for test of significance */
		value1[n] = sum_x[n] + sum_z[n];
		value2[n] = sq_sum_x[n] + sq_sum_z[n];
		value3[n] = ((sum_x[n]*sum_x[n])/(sequences*1.0)) + ((sum_z[n]*sum_z[n])/(ref_combs*1.0));
		value4[n] = (value1[n] * value1[n])/((sequences + ref_combs)*1.0);
		value5[n] = value2[n] - value4[n];
		value6[n] = value3[n] - value4[n];
		value7[n] = value5[n] - value6[n];
		value8[n] = value6[n]/((2 - 1)*1.0);
		value9[n] = value7[n]/((sequences + ref_combs - 2)*1.0);
		anova[n] = value8[n]/value9[n];
		/* obtaining the F[1,v2] critical value */
		(void)obtaining_f(&f, 1, (sequences + ref_combs - 2));
	        if ((anova[n] >= f) && (avg_x[n] > avg_z[n]))  {
	                chimera3[n] = 1;  /* significant P<0.05 */
	        }
	        else {
	                chimera3[n] = 0;
	        }


		/* Introduce here chimera3, for significance of the anova analysis */
	

		n = n + 1;
	}	

	/* If required, estimate breakpoint */


	/* Print results */
	(void)print_results(data_file_out, word, avg_x, sd_x, avg_z, sd_z, ds, chimera, chimera2, chimera3, fs, ts, anova, ref_combs, sequences, avg_nx, bases, name_seq, prediffs);
	
	/* close the output file */
	(void)fclose(data_file_out);

        return;

}


/****************************************************************
Removes the initial and final gaps to avoid biases due to incomplete sequences.
*****************************************************************/
void remove_extra_gaps( char seq[MAX_SEQS][MAX_BASES],
                	int bases[MAX_SEQS],
                	int sequences)
{

	int gotit;	  /* signals the begining of nucleotides in a sequence */
	int highest_gap[MAX_SEQS];/* gets the position of the highest gap at the initial part of the sequences */
	int lowest_gap[MAX_SEQS]; /* gets the lowest position of the gaps at the end of the sequences */
	char temp_seq[MAX_SEQS][MAX_BASES];  /* temporaly holds the sequences */
	int i, j;	  /* counters for FOR loops */
	int start_at;	  /* signals where to start the real sequence, without initial gaps */
	int end_at;	  /* signals where to ends the real sequence, without final gaps */
	int nucleotide;   /* counts number of nucleotide for temp_seq array */

	/* search for the initial filling gap */
	for (j=0; j<sequences+1; j++) {
		gotit = 0;
		highest_gap[j] = -1;
		for (i=0; i<bases[j]; i++) {
			if ((seq[j][i] == '-') && (gotit == 0)) {
				highest_gap[j] = i;
			}
			else if ((seq[j][i] != '-') && (gotit == 0)) {
				gotit = 1;
			}	
		}
	}
	start_at = -1;
	for (j=0; j<sequences+1; j++) {
		if (highest_gap[j] >= start_at) start_at = highest_gap[j] + 1;
	}

	/* search for the final filling gaps */
        for (j=0; j<sequences+1; j++) {
                gotit = 0;
                lowest_gap[j] = bases[j];
                for (i=bases[j]-1; i>=0; i--) {
                        if ((seq[j][i] == '-') && (gotit == 0)) {
                                lowest_gap[j] = i;
                        }
			else if ((seq[j][i] != '-') && (gotit == 0)) {
				gotit = 1;
			}
                }
        }
        end_at = bases[1];
        for (j=0; j<sequences+1; j++) {
                if (lowest_gap[j] <= end_at) end_at = lowest_gap[j] - 1;
	}

	/* remodel sequences removing initial and final gaps */
        for (j=0; j<sequences+1; j++) {
                nucleotide = 0;
                for (i=0; i<bases[j]; i++) {
			if ((i >= start_at) && (i <= end_at)) {
				temp_seq[j][nucleotide] = seq[j][i];
				nucleotide = nucleotide + 1;
			}
                }
		bases[j] = nucleotide;  /* update bases[] */
        }

	/* transfer temp_seq back to seq */
        for (j=0; j<sequences+1; j++) {
                for (i=0; i<bases[j]; i++) {
                        seq[j][i] = temp_seq[j][i];
                }
        }

        return;

}

/****************************************************************
Analyzes ambiguous bases searching for matching possibilities.
*****************************************************************/
void ambiguity( int     *different_base_ptr,
                char    base_a,
                char    base_b)
{
        int i; /* counter */
        int ok; /* ok=1 means equivalent base. Checks for degenerate bases */

	/* the char in base_a and base_b have been checked and they are different */
	if ((base_a == 'N') && (base_b != '-')) ok = 1;
	else if ((base_b == 'N') && (base_a != '-')) ok = 1;
	else if ((base_a == 'Y') && ((base_b == 'C') || (base_b == 'T'))) ok = 1;
        else if ((base_b == 'Y') && ((base_a == 'C') || (base_a == 'T'))) ok = 1;
        else if ((base_a == 'R') && ((base_b == 'G') || (base_b == 'A'))) ok = 1;
        else if ((base_b == 'R') && ((base_a == 'G') || (base_a == 'A'))) ok = 1;
        else if ((base_a == 'S') && ((base_b == 'C') || (base_b == 'G'))) ok = 1;
        else if ((base_b == 'S') && ((base_a == 'C') || (base_a == 'G'))) ok = 1;
        else if ((base_a == 'W') && ((base_b == 'T') || (base_b == 'A'))) ok = 1;
        else if ((base_b == 'W') && ((base_a == 'T') || (base_a == 'A'))) ok = 1;
        else if ((base_a == 'M') && ((base_b == 'A') || (base_b == 'C'))) ok = 1;
        else if ((base_b == 'M') && ((base_a == 'A') || (base_a == 'C'))) ok = 1;
        else if ((base_a == 'K') && ((base_b == 'T') || (base_b == 'G'))) ok = 1;
        else if ((base_b == 'K') && ((base_a == 'T') || (base_a == 'G'))) ok = 1;
        else if ((base_a == 'V') && ((base_b == 'C') || (base_b == 'A') || (base_b == 'G'))) ok = 1;
        else if ((base_b == 'V') && ((base_a == 'C') || (base_a == 'A') || (base_a == 'G'))) ok = 1;
        else if ((base_a == 'H') && ((base_b == 'T') || (base_b == 'A') || (base_b == 'C'))) ok = 1;
        else if ((base_b == 'H') && ((base_a == 'T') || (base_a == 'A') || (base_a == 'C'))) ok = 1;
        else if ((base_a == 'D') && ((base_b == 'T') || (base_b == 'A') || (base_b == 'G'))) ok = 1;
        else if ((base_b == 'D') && ((base_a == 'T') || (base_a == 'A') || (base_a == 'G'))) ok = 1;
        else if ((base_a == 'B') && ((base_b == 'C') || (base_b == 'T') || (base_b == 'G'))) ok = 1;
        else if ((base_b == 'B') && ((base_a == 'C') || (base_a == 'T') || (base_a == 'G'))) ok = 1;
	else ok = 0;  /* the bases are different and not equivalent */

        if (ok == 0) *different_base_ptr = 1;       /* bases are different and not equivalent */
	else if (ok == 1) *different_base_ptr = 0;  /* bases are equivalent */

        return;

}

/****************************************************************
 * Remove selected sequences
*****************************************************************/
void sequence_off(char seq[MAX_SEQS][MAX_BASES],
		  int bases[MAX_SEQS],
		  int *sequences_ptr,
		  int this_one[MAX_SEQS],
		  char name_seq[MAX_SEQS][MAX_LINE],
		  int prediffs[MAX_SEQS][MAX_SEQS])
{

	int l, m, o;	/* counters */
	int numseqs;  /* number of sequences */

	numseqs = *sequences_ptr;
	o = 0;
	for (l=1; l<numseqs+1; l++) {
		if (this_one[l] == 1) {
			/* Do nothing, this sequence should be removed */
		}
		else if (this_one[l] == 0) {
			o = o + 1;
			/* keep this sequence */	
			if (o < l) {
				for (m=0; m<=bases[l]; m++) {
					seq[o][m] = seq[l][m];
					(void)strcpy(name_seq[o], name_seq[l]);
					/* this is only considering differences with query seq. */
					prediffs[0][o] = prediffs[0][l]; 
				}
			} /* if */
		} /* else */
	}
	*sequences_ptr = o;

	return;

}

/****************************************************************
 Obtains the value of T-student for significance test
*****************************************************************/
void obtaining_t(float *t_ptr,
		 int df)
{

	int v;  /* degrees of freedom */

	v = df;

	/* t-student critical values for different degrees of freedom and alpha 0.1 in one-tail tests (equivalent to 0.05) */
	if (v > 120) *t_ptr = 1.645;
	else if (v > 60) *t_ptr = 1.658;
        else if (v > 40) *t_ptr = 1.671;
        else if (v > 30) *t_ptr = 1.684;
        else if (v > 29) *t_ptr = 1.697;
        else if (v > 28) *t_ptr = 1.699;
        else if (v > 27) *t_ptr = 1.701;
        else if (v > 26) *t_ptr = 1.703;
        else if (v > 25) *t_ptr = 1.706;
        else if (v > 24) *t_ptr = 1.708;
        else if (v > 23) *t_ptr = 1.711;
        else if (v > 22) *t_ptr = 1.714;
        else if (v > 21) *t_ptr = 1.717;
        else if (v > 20) *t_ptr = 1.721;
        else if (v > 19) *t_ptr = 1.725;
        else if (v > 18) *t_ptr = 1.729;
        else if (v > 17) *t_ptr = 1.734;
        else if (v > 16) *t_ptr = 1.740;
        else if (v > 15) *t_ptr = 1.746;
        else if (v > 14) *t_ptr = 1.753;
        else if (v > 13) *t_ptr = 1.761;
        else if (v > 12) *t_ptr = 1.771;
        else if (v > 11) *t_ptr = 1.782;
        else if (v > 10) *t_ptr = 1.796;
        else if (v > 9) *t_ptr = 1.812;
        else if (v > 8) *t_ptr = 1.833;
        else if (v > 7) *t_ptr = 1.860;
        else if (v > 6) *t_ptr = 1.895;
        else if (v > 5) *t_ptr = 1.943;
        else if (v > 4) *t_ptr = 2.015;
        else if (v > 3) *t_ptr = 2.132;
        else if (v > 2) *t_ptr = 2.353;
        else if (v > 1) *t_ptr = 2.920;
	else if (v <= 1) {
		(void)fprintf(stderr, "Two or more reference sequences are required.\n");
		exit(8);
	}

	return;

}

/********************************************************************
 *  Obtains the value of Ft for significance test if v1=1
 *  *****************************************************************/
void obtaining_f(float *f_ptr,
		 int v1,
		 int v2)
{

	/* this table contemplates only v1=1 i*/
	if (v1 != 1) (void)fprintf(stderr, "Ft values are only avaible for v1=1\n");
		
        /* F-Snedecor critical values for v1=1 and different degrees of freedom v2 and alpha 0.05 */
        if (v2 > 120) *f_ptr = 3.84;
        else if (v2 > 60) *f_ptr = 3.92;
        else if (v2 > 40) *f_ptr = 4.00;
        else if (v2 > 30) *f_ptr = 4.08;
        else if (v2 > 29) *f_ptr = 4.17;
        else if (v2 > 28) *f_ptr = 4.18;
        else if (v2 > 27) *f_ptr = 4.20;
        else if (v2 > 26) *f_ptr = 4.21;
        else if (v2 > 25) *f_ptr = 4.23;
        else if (v2 > 24) *f_ptr = 4.24;
        else if (v2 > 23) *f_ptr = 4.26;
        else if (v2 > 22) *f_ptr = 4.28;
        else if (v2 > 21) *f_ptr = 4.30;
        else if (v2 > 20) *f_ptr = 4.32;
        else if (v2 > 19) *f_ptr = 4.35;
        else if (v2 > 18) *f_ptr = 4.38;
        else if (v2 > 17) *f_ptr = 4.41;
        else if (v2 > 16) *f_ptr = 4.45;
        else if (v2 > 15) *f_ptr = 4.49;
        else if (v2 > 14) *f_ptr = 4.54;
        else if (v2 > 13) *f_ptr = 4.60;
        else if (v2 > 12) *f_ptr = 4.67;
        else if (v2 > 11) *f_ptr = 4.75;
        else if (v2 > 10) *f_ptr = 4.84;
        else if (v2 > 9) *f_ptr = 4.96;
        else if (v2 > 8) *f_ptr = 5.12;
        else if (v2 > 7) *f_ptr = 5.32;
        else if (v2 > 6) *f_ptr = 5.59;
        else if (v2 > 5) *f_ptr = 5.99;
        else if (v2 > 4) *f_ptr = 6.61;
        else if (v2 > 3) *f_ptr = 7.71;
        else if (v2 > 2) *f_ptr = 10.1;
        else if (v2 > 1) *f_ptr = 18.5;
        else if (v2 > 0) *f_ptr = 161;
	else if (v2 <= 0) {
                (void)fprintf(stderr, "Two or more reference sequences are required.\n");
                exit(8);
        }

        return;

}

/****************************************************************
Prints the header for the output file. 
*****************************************************************/
void print_header(FILE *data_file_out_ptr,
		  char *line_command)
{

	/* print heading for codon usage table */
	(void)fprintf(data_file_out_ptr, "-------------------------------------------------------------------------------- \n");
	(void)fprintf(data_file_out_ptr, "CCODE run: %s\n", line_command);
        (void)fprintf(data_file_out_ptr, "-------------------------------------------------------------------------------- \n");

	return;

}

/****************************************************************
Prints the results 
*****************************************************************/
void print_results(FILE  *data_file_out_ptr,
		   int word,
		   float avg_x[MAX_WORDS],
		   float sd_x[MAX_WORDS],
		   float avg_z[MAX_WORDS],
		   float sd_z[MAX_WORDS],
		   float ds[MAX_WORDS],
		   int chimera[MAX_WORDS],
		   int chimera2[MAX_WORDS],
		   int chimera3[MAX_WORDS],
		   float fs[MAX_WORDS],
		   float ts[MAX_WORDS],
		   float anova[MAX_WORDS], 
		   int ref_combs,
		   int sequences,
		   int avg_nx,
		   int bases[MAX_SEQS],
		   char name_seq[MAX_SEQS][MAX_LINE],
		   int prediffs[MAX_SEQS][MAX_SEQS])
{

	int m;	/* counter number of words */
	int l;	/* counter number of sequences */

	(void)fprintf(data_file_out_ptr, "Query sequence (Number of bases): %s (%d)\n", name_seq[0], bases[0]);
	(void)fprintf(data_file_out_ptr, "Reference sequences (Number of bases; Percent similarity with Query): \n");
	for (l=1; l<sequences+1; l++) {
	(void)fprintf(data_file_out_ptr, "   %s (%d; %.1f)\n", name_seq[l], bases[l], (100.0 - ((prediffs[0][l]*100.0)/(1.0*bases[l])))); 
	}
	(void)fprintf(data_file_out_ptr, "Word length = %d.\n", word);
	(void)fprintf(data_file_out_ptr, "============================================================================================\n");
/*	(void)fprintf(data_file_out_ptr, "Word\t AvgQ\t sdQ\t AvgR\t sdR\t ds\t Chimera\t Fs\n");
*/
	(void)fprintf(data_file_out_ptr, "Bases\t AvgQ (sdQ)\t AvgR (sdR)\t Ratio \tAnova[%d, %d]\n", (2 - 1), (sequences + ref_combs - 2));
	(void)fprintf(data_file_out_ptr, "============================================================================================\n");

	for (m=0; m<avg_nx; m++) {

	(void)fprintf(data_file_out_ptr, "%d \t %.2f (%.2f)\t %.2f (%.2f)\t %.2f", ((m+1)*word), avg_x[m], sd_x[m], avg_z[m], sd_z[m], ds[m]);

	/* print significance from confidence limit */
	if (chimera[m] == 0) (void)fprintf(data_file_out_ptr, " \t");
	else if (chimera[m] == 1) (void)fprintf(data_file_out_ptr, "*\t");

	/* print results for t-student and its significance */
/*	(void)fprintf(data_file_out_ptr, "%.3f", ts[m]);
	if (chimera2[m] == 0) (void)fprintf(data_file_out_ptr, " \t\t");
	else if (chimera2[m] == 1) (void)fprintf(data_file_out_ptr, "*\t\t");
*/

	(void)fprintf(data_file_out_ptr, "%.3f", anova[m]);
        if (chimera3[m] == 0) (void)fprintf(data_file_out_ptr, " \t\t");
	else if (chimera3[m] == 1) (void)fprintf(data_file_out_ptr, "*\t\t");
	
/*
	(void)fprintf(data_file_out_ptr, "%.3f", fs[m]);
*/

	(void)fprintf(data_file_out_ptr, "\n");

	}
	(void)fprintf(data_file_out_ptr, "============================================================================================\n");
	(void)fprintf(data_file_out_ptr, "Q -> Comparisons between query sequence and reference sequences.\n");
	(void)fprintf(data_file_out_ptr, "R -> Comparisons between reference sequences.\n");
/*	(void)fprintf(data_file_out_ptr, "Note: Chimera and Cross-over detections are only suggested events.\n");
*/
        (void)fprintf(data_file_out_ptr, "Significance at P<0.05 level is indicated by *.\n");
	(void)fprintf(data_file_out_ptr, "Number of pairwise comparisons between reference sequences = %d, and between query and reference sequences = %d.\n", ref_combs, sequences);
	(void)fprintf(data_file_out_ptr, "Results of ANOVA are given for the degrees of freedom between brackets.\n");

        return;
}

/****************************************************************
Print the copyright notice
*****************************************************************/
void print_cr(void)
{

	(void)fprintf(stdout, "\n \n \n **************************************************************\n \n");
	(void)fprintf(stdout, "  (C) Copyright. All rights reserved. Juan M. Gonzalez. 2004.\n");
	(void)fprintf(stdout, "  Permission to use, reproduce, and distribute this program\n");
	(void)fprintf(stdout, "  or its parts must be granted by the author.\n \n");
	(void)fprintf(stdout, "  Juan M. Gonzalez,\n");
	(void)fprintf(stdout, "  IRNAS-CSIC, P.O.Box 1052, 41080 - Sevilla, Spain.\n");
	(void)fprintf(stdout, "  jmgrau@irnase.csic.es \n \n");
	(void)fprintf(stdout, " **************************************************************\n \n \n");

	return;

}

/****************************************************************
Instructions for command-line arguments and program usage.
Prints the information and exits the program.
*****************************************************************/
void usage(char *program_name)
{

  (void)fprintf(stdout, "\n%s Chimera check and Cross-Over DEtection program.\n", program_name);
  (void)fprintf(stdout, "Sequences must be in fasta format. Query sequence should be first.\n");
  (void)fprintf(stdout, "\nUsage is %s -[options]<string or number>\n", program_name); 
  (void)fprintf(stdout, "Options are:\n");
  (void)fprintf(stdout, "  -I<input file>   Name of input file with sequences.\n");
  (void)fprintf(stdout, "  -O<output file>  Name for output file with the results.\n");
  (void)fprintf(stdout, "  -w<integer>      Word size as number of bases. Using -p is recommended.\n");
  (void)fprintf(stdout, "  -p<number>      Percentage of sequence length to be used for word size.\n");
  (void)fprintf(stdout, "  -f<integer>      Filter to apply.\n");
  (void)fprintf(stdout, "			0 = No filter. Default.\n");
  (void)fprintf(stdout, "			1 = Remove all gaps,\n");
  (void)fprintf(stdout, "			2 = Costume filter. (Not implemented yet. In todo list).\n");
  (void)fprintf(stdout, "  -M<number>       Maximum difference (%) accepted between two reference sequences.\n");
  (void)fprintf(stdout, "  -m<number>       Minimum difference (%) accepted between two reference sequences.\n");
  (void)fprintf(stdout, "  -r		    Do not check for reference sequences too similar or too distant.\n");


  (void)fprintf(stdout, "  -c               Avoids printing the copyright notice.\n");
  (void)fprintf(stdout, "  -h               Prints this usage information.\n");
  (void)fprintf(stdout, "Example:\n");
  (void)fprintf(stdout, "	%s -Iinputfile -Ooutputfile -p10 \n \n", program_name);

  exit(8);
}

